# Copyright (C) 2021 torque <torque@users.noreply.github.com>

# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.

# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# distutils: sources = ["deps/qrdec/src/binarize.c"]
# distutils: include_dirs = ["deps/qrdec/src"]
# distutils: extra_compile_args = ["-fdiagnostics-color=always"]

from . cimport binarize as bnz

cdef extern void qr_binarize(unsigned char *_img, int _width, int _height, int invert) nogil

cdef void binarize(bnz.uint8[::1] image, int width, int height, bint invert) nogil:
    # this is in-place manipulation
    qr_binarize(&image[0], width, height, invert)



from __future__ import annotations
from typing import Any

import Cython.Build
from setuptools.extension import Extension

def build(setup_kwargs: dict[str, Any]) -> None:
    cython_modules = Cython.Build.cythonize(
        [
            Extension(
                "deqr.quirc",
                sources=["deqr/quirc.pyx"]
            ),
            Extension(
                "deqr.qrdec",
                sources=["deqr/qrdec.pyx"]
            ),
            Extension(
                "deqr.binarize",
                sources=["deqr/binarize.pyx"]
            )
        ],
        language_level=3,
    )

    setup_kwargs.update(
        {
            "ext_modules": cython_modules,
            "zip_safe": False,
        }
    )



# Copyright (C) 2021 torque <torque@users.noreply.github.com>

# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.

# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# distutils: sources = ["deps/quirc/lib/decode.c", "deps/quirc/lib/identify.c", "deps/quirc/lib/quirc.c", "deps/quirc/lib/version_db.c"]
# distutils: include_dirs = ["deps/quirc/lib"]
# distutils: extra_compile_args = ["-fdiagnostics-color=always"]

from __future__ import annotations

cimport cython

import enum

from . cimport binarize as bnz, quircdecl
from . import datatypes, image

from libc.string cimport memcpy
from libc.stdio cimport printf, fopen, fwrite, fclose, FILE


cdef class QRDecoder:
    cdef quircdecl.quirc *_chndl

    def __cinit__(self):
        self._chndl = quircdecl.quirc_new()
        if self._chndl is NULL:
            raise MemoryError

    def __dealloc__(self):
        if self._chndl is not NULL:
            quircdecl.quirc_destroy(self._chndl)

    def resize(self, width: cython.int, height: cython.int):
        if quircdecl.quirc_resize(self._chndl, width, height) == -1:
            raise MemoryError

    def decode(
        self, image_data, binarize: bool = True, binarize_invert: bool = False
    ):
        cdef int idx = 0
        cdef quircdecl.quirc_code code
        cdef quircdecl.quirc_data data

        decoded: list[datatypes.QRCode] = []

        img = image.ImageLoader(image_data)

        if binarize:
            bnz.binarize(img.data, img.width, img.height, binarize_invert)

        for idx in range(self._set_image(img.data, img.width, img.height)):
            quircdecl.quirc_extract(self._chndl, idx, &code)
            quircdecl.quirc_decode(&code, &data)

            data_entries = (
                datatypes.QRCodeData(data.data_type, data.payload[:data.payload_len]),
            )

            center = self.compute_center_from_bounds(code.corners)

            decoded.append(
                datatypes.QRCode(
                    version=data.version,
                    ecc_level=data.ecc_level,
                    mask=data.mask,
                    data_entries=data_entries,
                    corners=tuple((c.x, c.y) for c in code.corners),
                    center=(center.x, center.y)
                )
            )

        return decoded

    cdef quircdecl.quirc_point compute_center_from_bounds(self, quircdecl.quirc_point corners[4]) nogil:
        cdef int divisor = 0
        cdef int lcoeff = 0
        cdef int rcoeff = 0

        divisor = (
            (corners[0].x - corners[2].x) * (corners[1].y - corners[3].y)
            - (corners[0].y - corners[2].y) * (corners[1].x - corners[3].x)
        )

        if divisor == 0:
            return quircdecl.quirc_point(-1, -1)
        else:
            lcoeff = corners[0].x * corners[2].y - corners[0].y * corners[2].x
            rcoeff = corners[1].x * corners[3].y - corners[1].y * corners[3].x

            return quircdecl.quirc_point(
                (
                    (corners[1].x - corners[3].x) * lcoeff
                    - (corners[0].x - corners[2].x) * rcoeff
                ) // divisor,
                (
                    (corners[1].y - corners[3].y) * lcoeff
                    - (corners[0].y - corners[2].y) * rcoeff
                ) // divisor
            )

    cdef int _set_image(self, quircdecl.uint8[::1] image, int width, int height) nogil:
        if quircdecl.quirc_resize(self._chndl, width, height) != quircdecl.QUIRC_SUCCESS:
            raise MemoryError("could not resize")

        cdef int bufsize = width*height

        cdef unsigned char *buffer = quircdecl.quirc_begin(self._chndl, NULL, NULL)
        memcpy(buffer, &image[0], bufsize)

        quircdecl.quirc_end(self._chndl)
        return quircdecl.quirc_count(self._chndl)



# Copyright (C) 2021 torque <torque@users.noreply.github.com>

# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.

# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

ctypedef unsigned char uint8
ctypedef unsigned int  uint32

cdef extern from "quirc.h":
    cdef struct quirc:
        pass

    const char *quirc_version() nogil

    quirc *quirc_new() nogil
    void quirc_destroy(quirc *q) nogil

    int quirc_resize(quirc *q, int width, int height) nogil

    uint8 *quirc_begin(quirc *q, int *width, int *height) nogil
    void quirc_end(quirc *q) nogil

    cdef struct quirc_point:
        int x
        int y

    ctypedef enum quirc_decode_error_t:
        QUIRC_SUCCESS = 0
        QUIRC_ERROR_INVALID_GRID_SIZE
        QUIRC_ERROR_INVALID_VERSION
        QUIRC_ERROR_FORMAT_ECC
        QUIRC_ERROR_DATA_ECC
        QUIRC_ERROR_UNKNOWN_DATA_TYPE
        QUIRC_ERROR_DATA_OVERFLOW
        QUIRC_ERROR_DATA_UNDERFLOW

    cdef enum:
        QUIRC_MAX_VERSION = 40
        QUIRC_MAX_GRID_SIZE = (QUIRC_MAX_VERSION * 4) + 17
        QUIRC_MAX_BITMAP = ((QUIRC_MAX_GRID_SIZE * QUIRC_MAX_GRID_SIZE) + 7) // 8
        QUIRC_MAX_PAYLOAD = 8896

    cdef enum:
        QUIRC_ECC_LEVEL_M = 0
        QUIRC_ECC_LEVEL_L = 1
        QUIRC_ECC_LEVEL_H = 2
        QUIRC_ECC_LEVEL_Q = 3

    cdef enum:
        QUIRC_DATA_TYPE_NUMERIC = 1
        QUIRC_DATA_TYPE_ALPHA = 2
        QUIRC_DATA_TYPE_BYTE = 4
        QUIRC_DATA_TYPE_KANJI = 8

    cdef enum:
        QUIRC_ECI_ISO_8859_1 = 1
        QUIRC_ECI_IBM437 = 2
        QUIRC_ECI_ISO_8859_2 = 4
        QUIRC_ECI_ISO_8859_3 = 5
        QUIRC_ECI_ISO_8859_4 = 6
        QUIRC_ECI_ISO_8859_5 = 7
        QUIRC_ECI_ISO_8859_6 = 8
        QUIRC_ECI_ISO_8859_7 = 9
        QUIRC_ECI_ISO_8859_8 = 10
        QUIRC_ECI_ISO_8859_9 = 11
        QUIRC_ECI_WINDOWS_874 = 13
        QUIRC_ECI_ISO_8859_13 = 15
        QUIRC_ECI_ISO_8859_15 = 17
        QUIRC_ECI_SHIFT_JIS = 20
        QUIRC_ECI_UTF_8 = 26

    cdef struct quirc_code:
        quirc_point corners[4]
        int         size
        uint8       cell_bitmap[QUIRC_MAX_BITMAP]

    # /* This structure holds the decoded QR-code data */
    cdef struct quirc_data:
        int    version
        int    ecc_level
        int    mask
        int    data_type
        uint8  payload[QUIRC_MAX_PAYLOAD]
        int    payload_len
        uint32 eci

    int quirc_count(const quirc *q) nogil
    void quirc_extract(const quirc *q, int index, quirc_code *code) nogil
    quirc_decode_error_t quirc_decode(const quirc_code *code, quirc_data *data) nogil
    void quirc_flip(quirc_code *code) nogil



from __future__ import annotations
from typing import Any

from . import binarize


class ImageLoader:
    width: int
    height: int
    data: Any

    def __init__(self, data: Any):
        try:
            if "numpy.ndarray" in str(type(data)):
                self.set_data_from_numpy_ndarray(data)
            elif "cv2.UMat" in str(type(data)):
                self.set_data_from_numpy_ndarray(data.get())
            elif "PIL.Image" in str(type(data)):
                self.set_data_from_pil_image(data)
            elif isinstance(data, tuple):
                self.set_data_from_tuple(data)
            else:
                raise Exception
        except (ValueError, TypeError):
            raise
        except Exception:
            raise TypeError(f"data {type(data)} is an incompatible type.")

    def set_data_from_numpy_ndarray(self, data):
        if data.dtype.name != "uint8":
            # data.astype is not really usable here as it just does blind casts,
            # which doesn't scale properly from larger types. At the same time,
            # it's not possible to really guess what the appropriate range of
            # the input data is for useful conversion
            raise ValueError(f"np.ndarray dtype {data.dtype} must be uint8")
        if data.ndim == 3:
            # try to convert sanely by averaging values
            data = data.mean(axis=2).astype(data.dtype)
        elif data.ndim != 2:
            raise ValueError(f"np.ndarray doesn't look like a 2D image {data.shape}")

        self.data = data.ravel(order="C")
        self.height, self.width = data.shape

    def set_data_from_pil_image(self, data):
        if data.mode != "L":
            data = data.convert("L")

        self.data = data.tobytes()
        self.width, self.height = data.size

    def set_data_from_tuple(self, data):
        if (
            len(data) != 2
            or not isinstance(data[0], bytes)
            or len(data[1]) != 2
            or not isinstance(data[1][0], int)
            or not isinstance(data[1][1], int)
        ):
            raise TypeError("Data tuple is not of format (bytes, (int, int))")

        if len(data[0]) != (data[1][0] * data[1][1]):
            raise ValueError(
                f"Wrong amount of image data {len(data[0])} for "
                f"{data[1][0]} by {data[1][1]} image"
            )

        (self.data, (self.width, self.height)) = data



# Copyright (C) 2021 torque <torque@users.noreply.github.com>

# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.

# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

ctypedef unsigned char uint8
ctypedef unsigned int  uint32

cdef extern from "qrcode.h":
    ctypedef int qr_point[2]

    cdef struct qr_reader:
        pass

    cdef struct qr_code_data_entry_raw_data:
        unsigned char *buf
        int len

    cdef struct qr_code_data_entry_sa_header:
        unsigned char sa_index
        unsigned char sa_size
        unsigned char sa_parity

    cdef union qr_code_data_entry_payload:
        qr_code_data_entry_raw_data data
        unsigned int eci
        int ai
        qr_code_data_entry_sa_header sa

    cdef struct qr_code_data_entry:
        qr_mode mode
        qr_code_data_entry_payload payload

    cdef struct qr_code_data:
        qr_code_data_entry *entries;
        int nentries;
        unsigned char version;
        unsigned char ecc_level;
        unsigned char mask;
        unsigned char sa_index;
        unsigned char sa_size;
        unsigned char sa_parity;
        unsigned char self_parity;
        qr_point bbox[4];
        qr_point center;

    cdef struct qr_code_data_list:
        qr_code_data *qrdata
        int nqrdata
        int cqrdata

    ctypedef enum qr_mode:
        QR_MODE_NUM = 1
        QR_MODE_ALNUM
        QR_MODE_STRUCT
        QR_MODE_BYTE
        QR_MODE_FNC1_1ST
        QR_MODE_ECI = 7
        QR_MODE_KANJI
        QR_MODE_FNC1_2ND

    ctypedef enum qr_eci_encoding:
        QR_ECI_GLI0 = 0
        QR_ECI_GLI1
        QR_ECI_CP437
        QR_ECI_ISO8859_1
        QR_ECI_ISO8859_2
        QR_ECI_ISO8859_3
        QR_ECI_ISO8859_4
        QR_ECI_ISO8859_5
        QR_ECI_ISO8859_6
        QR_ECI_ISO8859_7
        QR_ECI_ISO8859_8
        QR_ECI_ISO8859_9
        QR_ECI_ISO8859_10
        QR_ECI_ISO8859_11
        QR_ECI_ISO8859_13 = QR_ECI_ISO8859_11 + 2
        QR_ECI_ISO8859_14
        QR_ECI_ISO8859_15
        QR_ECI_ISO8859_16
        QR_ECI_SJIS = 20
        QR_ECI_UTF8 = 26

    uint8 QR_MODE_HAS_DATA(qr_mode mode)
    qr_reader * qr_reader_alloc() nogil
    void qr_reader_free(qr_reader *_reader) nogil
    void qr_code_data_list_init(qr_code_data_list *_qrlist) nogil
    void qr_code_data_list_clear(qr_code_data_list *_qrlist) nogil

    int qr_reader_locate(
        qr_reader *_reader,
        qr_code_data_list *_qrlist,
        const unsigned char *_img,
        int _width,int _height
    ) nogil


    int qr_reader_extract_text(
        qr_reader *_reader,
        const unsigned char *_img,
        int _width,
        int _height,
        char ***_text,
        int _allow_partial_sa
    ) nogil
    void qr_text_list_free(char **_text, int _ntext) nogil



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['deqr']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'deqr',
    'version': '0.1.2',
    'description': 'qr code decoding library',
    'long_description': '## deqr\n\nA python library for decoding QR codes. Implemented as a cython wrapper around\ntwo different QR code decoding backends (quirc and qrdec).\n',
    'author': 'torque',
    'author_email': 'torque@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/torque/deqr',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



from __future__ import annotations

import enum


class QREccLevel(enum.IntEnum):
    M = 0
    L = 1
    H = 2
    Q = 3


class QRDataType(enum.IntEnum):
    NUMERIC = 1
    ALPHANUMERIC = 2
    BYTE = 4
    KANJI = 8


class QRCodeData:
    type: common.QRDataType
    data: bytes

    def __init__(self, type, data):
        self.type = QRDataType(type)
        self.data = data

    def __str__(self):
        return "QrCodeData(type=%s, data=<%s>)" % (
            self.type,
            self.data.hex(" ").upper(),
        )

    def __repr__(self):
        return str(self)


class QRCode:
    """
    Information about a physical QR code that has been decoded.
    """

    version: int
    ecc_level: QREccLevel
    mask: int

    data_entries: tuple[QRCodeData]

    center: tuple[int, int]
    corners: tuple[tuple[int, int], ...]

    def __init__(self, version, ecc_level, mask, data_entries, corners, center):
        self.version = version
        self.ecc_level = QREccLevel(ecc_level)
        self.mask = mask
        self.data_entries = data_entries

        self.corners = corners
        self.center = center

    def __str__(self):
        return (
            "QRCode(version=%d, ecc_level=%s, mask=%d, data_entries=%s, "
            "corners=%s, center=%s)"
        ) % (
            self.version,
            self.ecc_level,
            self.mask,
            self.data_entries,
            self.corners,
            self.center
        )

    def __repr__(self):
        return str(self)



# Copyright (C) 2021 torque <torque@users.noreply.github.com>

# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.

# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# distutils: sources = ["deps/qrdec/src/bch15_5.c", "deps/qrdec/src/isaac.c", "deps/qrdec/src/qrdec.c", "deps/qrdec/src/rs.c", "deps/qrdec/src/util.c"]
# distutils: include_dirs = ["deps/qrdec/src"]
# distutils: extra_compile_args = ["-fdiagnostics-color=always"]

from __future__ import annotations

cimport cython

import enum

from . cimport binarize as bnz, qrdecdecl
from . import datatypes, image

from libc.string cimport memcpy
from libc.stdio cimport printf, fopen, fwrite, fclose, FILE


QREccLevelMap = {
    0: datatypes.QREccLevel.L,
    1: datatypes.QREccLevel.M,
    2: datatypes.QREccLevel.Q,
    3: datatypes.QREccLevel.H,
}

cdef class QRDecoder:
    cdef qrdecdecl.qr_reader *_chndl

    def __cinit__(self):
        self._chndl = qrdecdecl.qr_reader_alloc()
        if self._chndl is NULL:
            raise MemoryError

    def __dealloc__(self):
        if self._chndl is not NULL:
            qrdecdecl.qr_reader_free(self._chndl)

    def decode(
        self, image_data, binarize: bool = True, binarize_invert: bool = True
    ):
        cdef qrdecdecl.qr_code_data_list qrlist

        img = image.ImageLoader(image_data)

        qrdecdecl.qr_code_data_list_init(&qrlist)

        cdef qrdecdecl.uint8[::1] imagebytes = img.data
        cdef int idx = 0
        cdef qrdecdecl.qr_code_data *code
        decoded: list[datatypes.QRCode] = []

        if binarize:
            bnz.binarize(img.data, img.width, img.height, binarize_invert)

        try:
            for idx in range(
                qrdecdecl.qr_reader_locate(
                    self._chndl, &qrlist, &imagebytes[0], img.width, img.height
                )
            ):
                code = qrlist.qrdata + idx
                corners = tuple((corner[0], corner[1]) for corner in code.bbox)
                data_entries = tuple(
                    datatypes.QRCodeData(
                        entry.mode,
                        entry.payload.data.buf[:entry.payload.data.len]
                    )
                    for entry in code.entries[:code.nentries]
                    if qrdecdecl.QR_MODE_HAS_DATA(entry.mode)
                )


                decoded.append(
                    datatypes.QRCode(
                        version=code.version,
                        ecc_level=QREccLevelMap[code.ecc_level],
                        mask=code.mask,
                        data_entries=data_entries,
                        corners=corners,
                        center=(code.center[0], code.center[1])
                    )
                )
        finally:
            qrdecdecl.qr_code_data_list_clear(&qrlist)

        return decoded
